/* gc-minor.c */

/* code du gc mineur. (stop&copy sur la generation) */

#include "mlvalues.h"
#include "gc.h"
#include "gc_aux.h"
#include "mem_misc.h"
#include "free_list.h"
#include "roots.h"
#include "fail.h"

extern address heap_start, heap_end;
extern address young_end, young_start;

/**** tables des references inverses ****/
/* depuis les vieux */
value **ref_index, **cur_ref;
value *ref_table [REF_TABLE_SIZE];
value **ref_table_end = &ref_table [REF_TABLE_SIZE];

value extra_root;
int minor_gc_is_running;
int out_of_memory;
int major_gc_did_occur;

address *copied_young;
address *scan_young, *copy_young;

value alloc_in_free_list ();

static value dummy = 0;

/* allocation dans la zone des vieux. */
/* base doit etre une racine deja decrite dans les racines du GC majeur */
#define BIG_ALLOC(result, len, save, base)                                    \
/* address result; blocksize len; address save; value base; */                \
  { Zassert ((len) <= MAX_SIZE);                                              \
    result = hp;                                                              \
    hp += (len + 1);                                                          \
    if (hp >= block_end){                                                     \
      value base_save = (base);                                               \
      hp -= (len + 1);                                                        \
      major_gc_did_occur = 0;                                                 \
      result = ADDRESS (alloc_in_free_list (len + 1));                        \
      if ((base) == MLINT (0L) && IS_IN_HEAP (save)){                         \
        /* il s'agissait d'un pointeur inverse, donc le gc majeur l'a remis   \
	   dans la table des pointeurs inverses : on neutralise la mise a     \
	   jour si le gc majeur s'est declenche, car l'objet a peut-etre      \
           bouge */                                                           \
	if (major_gc_did_occur){                                              \
	  (save) = &dummy;                                                    \
	}                                                                     \
      }else{                                                                  \
	(save) += ADDRESS (base) - ADDRESS (base_save);                       \
      }                                                                       \
    }                                                                         \
  }

/* copier un bloc ou mettre a jour une reference */
#define COPY_TO_OLD(adr, base)                                                \
/* value *adr; value base; */                                                 \
  { register address co_block = ADDRESS (*(adr));                             \
    /* adr est l'adresse du pointeur a mettre a jour,                         \
       et co_block est l'adresse de l'objet a copier */                       \
                                                                              \
    Zassert (IS_BLOCK (*(adr)));                                              \
    Zassert (IS_YOUNG (*(adr)));                                              \
                                                                              \
    if (IS_FORWARDED (OBJECT (co_block))){                                    \
      /* bloc deja copie. */                                                  \
      *(adr) = FORWARD_POINTER (OBJECT (co_block));                           \
    }else{                                                                    \
      /* bloc a copier */                                                     \
      register address co_p;                                                  \
      register address co_p1 = co_block; /* il faut sauver l'ancienne adresse \
					    de co_block car le gc majeur peut \
					    avoir modifie *(adr) */           \
      register value co_q;                                                    \
      register blocksize i = H_SIZE (co_block);                               \
                                                                              \
      BIG_ALLOC (co_p, i, (adr), (base));                                     \
      co_q = OBJECT (co_p);                                                   \
                                                                              \
      if (H_TAG (co_block) < NO_SCAN_TAG){                                    \
	*copy_young++ = co_p;	/* mettre le bloc dans la file */             \
      }                                                                       \
                                                                              \
      *(header *) co_p = *(header *) co_block;       /* copier le header */   \
      co_p += SIZE_HEADER; co_block += SIZE_HEADER;                           \
      *co_p++ = *co_block++;			/* copier le premier champ */ \
      if (i -= 1){                                                            \
	do{                                                                   \
	  *co_p++ = *co_block++;                                              \
	}while (--i);                                                         \
      }                                                                       \
      PUT_FORWARD_FLAG (OBJECT (co_p1));                                      \
      FORWARD_POINTER (OBJECT (co_p1)) = co_q;                                \
      *(adr) = co_q;                                                          \
    }                                                                         \
  }

/* copier tout ce qui est accessible a partir d'une racine. */

void copy_from (root)
     value *root;
{
  Zassert (IS_BLOCK (*root));
  Zassert (IS_YOUNG (*root));

  /* extra_root pour le cas ou le gc majeur coupe la pile. root doit rester
     vivant.
  */
  extra_root = *root;
  COPY_TO_OLD (root, MLINT (0L));
  while (scan_young < copy_young){
    register address p = *scan_young;
    register blocksize i = H_SIZE (p);
    
    Zassert (IS_IN_HEAP (OBJECT (p)));
    Zassert (H_TAG (p) < NO_SCAN_TAG);
    Zassert (i >= 1);
    
    p += SIZE_HEADER;					 /* header */
    
    if (IS_YOUNG (*p) && IS_BLOCK (*p)){		 /* premier champ */
      COPY_TO_OLD (p, OBJECT (*scan_young));
    }
    p++;
    
    if (i -= 1){
      do{
	if (IS_YOUNG (*p) && IS_BLOCK (*p)){
	  COPY_TO_OLD (p, OBJECT (*scan_young));
	}
	p++;
      }while (--i);
    }
    scan_young++;
  }
}

void minor_gc ()
{
  /* out_of_memory est mis a 1 par le GC majeur s'il a ete oblige
     de couper la pile pour faire de la place (cad de commencer un raise.)
     Si c'est le cas, il faut remettre le pointeur d'exception en place
     et appeler raise aussitot que le tas est dans un etat convenable
  */
  minor_gc_is_running = 1;
  out_of_memory = 0;

  /* les blocs alloues temporairement sont desalloues au moindre GC, et
     ce n'est pas une menace en l'air !
  */
  temp_free ();

  copy_young = scan_young = copied_young;
  
  /* parcours de la table des references inverses. */
  for (cur_ref = ref_table; cur_ref < ref_index; cur_ref++){
    if (IS_BLOCK (**cur_ref) && IS_YOUNG (**cur_ref)){
      copy_from (*cur_ref);
    }
  }
  
  /* parcours des racines intrinseques */
  ROOTS_FOR_MINOR (copy_from);

#ifdef DEBUG
  { blocksize i;
    for (i = 0; i < GENERATION_SIZE; i++){
      young_start [i] = gc_debug_random ();
    }
  }
#endif

  /* pour eviter de troubler le gc majeur s'il se declenche entre
     deux gc mineurs */
  scan_young = copy_young = copied_young;
  extra_root = NULL;
  minor_gc_is_running = 0;

  if (out_of_memory){
    /* on peut enfin declencher l'exception. */
    system_failwith_out_of_memory ();
  }
}

void init_minor_gc ()
{
  copied_young =
    (address *) stat_alloc((young_end - young_start) / 2 * sizeof(address));
  /* pour eviter de troubler le gc majeur s'il se declenche avant le premier
     gc mineur (ca arrive) */
  scan_young = copy_young = copied_young;
  extra_root = NULL;
  minor_gc_is_running = 0;
}
