#open "obj";;
#open "hash";;
#open "lex_run";;

let lex_val = ref(Repr 1);;
let comment_depth = ref 0;;

let keyword_table = make_vect 31 ([] : (string * int) list)
and infix_table = make_vect 17  ([] : string list);;

let find_keyword s =
  match keyword_table.(hash_univ_param 1 1 s  mod  31) with
    [] -> raise (failure "find")
  | (s1,tok1)::rest1 ->
      if s = s1 then tok1 else match rest1 with
        [] -> raise (failure "find")
      | (s2,tok2)::rest2 ->
          if s = s2 then tok2 else match rest2 with
            [] -> raise (failure "find")
          | (s3,tok3)::rest3 ->
              if s = s3 then tok3 else  assoc s rest3
;;

let is_infix s =
  match infix_table.(hash_univ_param 1 1 s  mod  17) with
    [] -> false
  | [s1] -> s = s1
  | [s1;s2] -> s = s1 or s = s2
  | l -> mem s l
;;

let initial_string_buffer = create_string 256;;
let string_buff = ref initial_string_buffer;;
let string_index = ref 0;;
let string_start = ref 0;;

let reset_string_buffer () =
  string_buff := initial_string_buffer;
  string_index := 0;
  string_start := !lex_start_pos;
  ()
;;

let store_string_char c =
  (if !string_index >= string_length (!string_buff) then
    let new_buff = create_string (string_length (!string_buff) * 2) in
      blit_string new_buff 0 (!string_buff) 0 (string_length (!string_buff));
      string_buff := new_buff;
      ());
  set_nth_char (!string_buff) (!string_index) c;
  incr string_index
;;

let get_stored_string () =
  lex_start_pos := !string_start;
  let s = sub_string (!string_buff) 0 (!string_index) in
    string_buff := initial_string_buffer;
    s
;;

let char_for_backslash = function
    `n` -> config__newline_char
  | `r` -> config__return_char
  | `t` -> `\009`
  | `b` -> `\008`
  | c   -> c
;;

let char_for_decimal_code i =
  char_of_int(100 * (int_of_char(get_lexeme_char i) - 48) +
               10 * (int_of_char(get_lexeme_char (i+1)) - 48) +
                    (int_of_char(get_lexeme_char (i+2)) - 48))
;;

